from typing import AsyncContextManager, ContextManager

import asyncio
from functools import partial
from unittest.mock import ANY, AsyncMock, MagicMock, Mock, call, patch

import pytest

from mode import Service
from mode.services import Diag, ServiceTask, WaitResult
from mode.utils.logging import get_logger


class S(Service):
    def __post_init__(self):
        self.on_started_log = Mock()
        self.on_stopped_log = Mock()
        self.on_shutdown_log = Mock()

    async def on_start(self):
        self.on_started_log()

    async def on_stop(self):
        self.on_stopped_log()

    async def on_shutdown(self):
        self.on_shutdown_log()


class test_Diag:
    @pytest.fixture()
    def diag(self):
        service = Mock()
        return Diag(service)

    def test_set_unset_flag(self, *, diag):
        diag.set_flag("FOO")
        assert "FOO" in diag.flags
        assert diag.last_transition["FOO"]
        diag.unset_flag("FOO")
        assert "FOO" not in diag.flags


class test_ServiceTask:
    @pytest.fixture()
    def task(self):
        fun = AsyncMock()
        return ServiceTask(fun)

    @pytest.mark.asyncio
    async def test_call(self, *, task):
        obj = Mock()
        ret = await task(obj)
        assert ret is task.fun.return_value
        task.fun.assert_awaited_once_with(obj)

    def test_repr(self, *, task):
        assert repr(task) == repr(task.fun)


@pytest.mark.asyncio
async def test_start_stop():
    s = S()
    assert s.state == "init"
    assert await s.maybe_start()
    assert not await s.maybe_start()
    assert s.state == "running"
    s.on_started_log.assert_called_with()
    await s.stop()
    s.on_stopped_log.assert_called_with()
    s.on_shutdown_log.assert_called_with()
    assert s.state == "stopping"


def test_state_stopped():
    s = S()
    s._started.set()
    s._stopped.set()
    s._shutdown.set()
    assert s.state == "shutdown"


def test_should_stop_returns_true_if_crashed():
    s = S()
    s._crashed.set()
    assert s.should_stop


@pytest.mark.asyncio
async def test_aenter():
    s = S()
    async with s:
        s.on_started_log.assert_called_with()
    s.on_stopped_log.assert_called_with()
    s.on_shutdown_log.assert_called_with()


@pytest.mark.asyncio
async def test_interface():
    s = Service()
    s.__post_init__()
    await s.on_start()
    await s.on_stop()
    await s.on_shutdown()


def test_repr():
    assert repr(S())


@pytest.mark.asyncio
async def test_subclass_can_override_Service_task():
    class ATaskService(Service):
        values = []

        def __post_init__(self):
            self.event = asyncio.Event()

        @Service.task
        async def _background_task(self):
            self.values.append(1)
            self.event.set()

    class BTaskService(ATaskService):
        @Service.task
        async def _background_task(self):
            self.values.append(2)
            self.event.set()

    async with BTaskService() as service:
        await service.event.wait()

    assert ATaskService.values == [2]


class test_Service:
    @pytest.fixture()
    def service(self):
        return S()

    @pytest.mark.asyncio
    async def test_from_awaitable(self, *, service):
        foo = Mock()

        async def mywait():
            foo()

        s = service.from_awaitable(mywait())
        async with s:
            pass
        foo.assert_called_once_with()
        await s.on_stop()

    @pytest.mark.asyncio
    async def test_from_awaitable__sleeping_is_cancelled(self, *, service):
        foo = Mock()

        async def mywait():
            foo()
            asyncio.ensure_future(s.stop())
            await asyncio.sleep(10.0)

        s = service.from_awaitable(mywait())
        async with s:
            await s.stop()
            await asyncio.sleep(0)
            await asyncio.sleep(0)
        foo.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_from_awaitable__raises_cancel(self, *, service):
        foo = Mock()

        async def mywait():
            foo()
            raise asyncio.CancelledError()

        s = service.from_awaitable(mywait())
        with pytest.raises(asyncio.CancelledError):
            async with s:
                pass
        foo.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_timer(self):
        m = Mock()

        class Foo(Service):
            @Service.timer(1.0)
            async def foo(self):
                m()
                if m.call_count >= 3:
                    self._stopped.set()

        async def itertimer(*args, **kwargs):
            yield 0.1
            yield 0.2
            yield 0.3
            yield 0.4

        foo = Foo()
        foo.sleep = AsyncMock()
        foo.itertimer = itertimer
        async with foo:
            pass
        assert m.call_count == 4

    @pytest.mark.asyncio
    async def test_timer__not_to_wait_twice(self) -> None:
        m = Mock()

        class Foo(Service):
            @Service.timer(0.1)
            async def foo(self) -> None:
                m()
                if m.call_count >= 3:
                    self._stopped.set()

        async def itertimer(self, t, *args, **kwargs):
            for i in range(4):
                await self.sleep(t)
                yield

        foo = Foo()
        foo.sleep = AsyncMock()
        foo.itertimer = partial(itertimer, foo)
        async with foo:
            pass

        assert foo.sleep.call_count == 4
        assert m.call_count == 4

    @pytest.mark.asyncio
    async def test_timer__exec_first(self) -> None:
        m = Mock()

        class Foo(Service):
            @Service.timer(1.0, exec_first=True)
            async def foo(self) -> None:
                m()
                if m.call_count >= 3:
                    self._stopped.set()

        async def itertimer(*args, **kwargs):
            yield 0.1
            yield 0.2
            yield 0.3

        foo = Foo()
        foo.sleep = AsyncMock()
        foo.itertimer = itertimer
        async with foo:
            pass
        assert m.call_count == 4

    @pytest.mark.asyncio
    async def test_crontab(self):
        m = Mock()

        with patch("mode.services.secs_for_next") as secs_for_next:
            secs_for_next.secs_for_next.return_value = 0.1

            class Foo(Service):
                @Service.crontab("* * * * *")
                async def foo(self):
                    m()
                    self._stopped.set()

            foo = Foo()
            foo.sleep = AsyncMock()
            async with foo:
                await asyncio.sleep(0)

            m.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_transitions_to(self, *, service):
        @service.transitions_to("FLAG")
        async def foo(self, arg, kw=1):
            assert "FLAG" in service.diag.flags
            assert arg == 1
            assert kw == 2

        await foo(service, 1, kw=2)
        assert "FLAG" not in service.diag.flags

    @pytest.mark.asyncio
    async def test_transition_with(self, *, service):
        called = Mock()

        async def outer():
            assert "FLAG" in service.diag.flags
            called()

        await service.transition_with("FLAG", outer())
        called.assert_called_once_with()

    def test_add_dependency__no_beacon(self, *, service):
        m = Mock()
        m.beacon = None
        service.add_dependency(m)

    @pytest.mark.asyncio
    async def test_add_runtime_dependency__when_started(self, *, service):
        s2 = Mock(maybe_start=AsyncMock())
        service.add_dependency = Mock()
        service._started.set()
        await service.add_runtime_dependency(s2)
        service.add_dependency.assert_called_once_with(s2)
        s2.maybe_start.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_add_runtime_dependency__not_started(self, *, service):
        s2 = Mock(maybe_start=AsyncMock())
        service.add_dependency = Mock()
        service._started.clear()
        await service.add_runtime_dependency(s2)
        service.add_dependency.assert_called_once_with(s2)
        s2.maybe_start.assert_not_called()

    @pytest.mark.asyncio
    async def test_remove_dependency(self, *, service):
        s2 = Mock(stop=AsyncMock())
        service.add_dependency(s2)
        service._started.set()
        await service.remove_dependency(s2)
        s2.stop.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_remove_dependency__no_beacon(self, *, service):
        s2 = Mock(stop=AsyncMock())
        s2.beacon = None
        service.add_dependency(s2)
        await service.remove_dependency(s2)

    @pytest.mark.asyncio
    async def test_add_async_context__non_async(self, *, service):
        class Cx(ContextManager):
            def __exit__(self, *args):
                return None

        assert isinstance(Cx(), ContextManager)

        with pytest.raises(TypeError):
            await service.add_async_context(Cx())

    @pytest.mark.asyncio
    async def test_add_async_context__not_context(self, *, service):
        with pytest.raises(TypeError):
            await service.add_async_context(object())

    def test_add_context__is_async(self, *, service):
        class Cx(AsyncContextManager):
            async def __aexit__(self, *args):
                return None

        assert isinstance(Cx(), AsyncContextManager)

        with pytest.raises(TypeError):
            service.add_context(Cx())

    def test_add_context__not_context(self, *, service):
        with pytest.raises(TypeError):
            service.add_context(object())

    @pytest.mark.asyncio
    async def test__wait_stopped(self, *, service):
        service._stopped = Mock()
        service._crashed = Mock()

        with (
            patch("asyncio.wait", AsyncMock()) as wait,
            patch("asyncio.ensure_future", Mock()) as ensure_future,
        ):
            f1 = Mock()
            f2 = Mock()
            f3 = Mock()
            done, pending = wait.return_value = ((f1,), (f2, f3))

            await service._wait_stopped(timeout=1.0)

            service._stopped.wait.assert_called_once()
            service._crashed.wait.assert_called_once()

            ensure_future.assert_has_calls(
                [
                    call(service._stopped.wait.return_value, loop=service.loop),
                    call(service._crashed.wait.return_value, loop=service.loop),
                ]
            )

            stopped = ensure_future.return_value
            crashed = ensure_future.return_value
            wait.assert_called_once_with(
                [stopped, crashed],
                return_when=asyncio.FIRST_COMPLETED,
                timeout=1.0,
            )

            for fut in done:
                fut.result.assert_called_once_with()
            for fut in pending:
                fut.cancel.assert_called_once_with()

        service._crashed.is_set.assert_called_once_with()

    @pytest.mark.asyncio
    async def test__actually_start__add_dependencies(self, *, service):
        s1 = Mock()
        s2 = Mock()
        self._mock_for_start(service, init_deps=[s1, s2])
        service.on_first_start.side_effect = service._stopped.set
        await service._actually_start()
        service.add_dependency.assert_has_calls([call(s1), call(s2)])

    @pytest.mark.asyncio
    async def test__actually_start__second_stop(self, *, service):
        s1 = Mock()
        self._mock_for_start(
            service,
            init_deps=[s1],
            on_async_enter=service._stopped.set,
        )
        service.restart_count = 1
        await service._actually_start()

    @pytest.mark.asyncio
    async def test__actually_start__third_stop(self, *, service):
        s1 = Mock()
        self._mock_for_start(
            service,
            init_deps=[s1],
        )
        service.on_start.side_effect = service._stopped.set
        service.restart_count = 1
        await service._actually_start()

    @pytest.mark.asyncio
    async def test__actually_start__rasies_exc(self, *, service):
        s1 = Mock()
        self._mock_for_start(
            service,
            init_deps=[s1],
        )
        service.on_start.side_effect = KeyError("foo")
        service.restart_count = 1
        with pytest.raises(KeyError):
            await service._actually_start()

    @pytest.mark.asyncio
    async def test__actually_start__fourth_stop(self, *, service):
        s1 = Mock(maybe_start=AsyncMock())
        self._mock_for_start(
            service,
        )
        service._children = [s1]
        s1.maybe_start.side_effect = service._stopped.set
        service.restart_count = 1
        await service._actually_start()

    @pytest.mark.asyncio
    async def test__actually_start__child_is_None(self, *, service):
        s1 = Mock(maybe_start=AsyncMock())
        self._mock_for_start(
            service,
        )
        service._children = [None, s1]
        s1.maybe_start.side_effect = service._stopped.set
        service.restart_count = 1
        await service._actually_start()

    def _mock_for_start(
        self, service, init_deps=None, tasks=None, children=None, on_async_enter=None
    ):
        service.on_init_dependencies = Mock(return_value=init_deps or [])
        service.add_dependency = Mock()
        service.on_first_start = AsyncMock()
        service.exit_stack = MagicMock()
        service.async_exit_stack = AsyncMock(
            side_effect=on_async_enter,
        )
        service.on_start = AsyncMock()
        service._get_tasks = Mock(return_value=tasks or [])
        service._children = children or []

    @pytest.mark.asyncio
    async def test_join_services(self, *, service):
        s1 = Mock(maybe_start=AsyncMock(), stop=AsyncMock())
        s2 = Mock(maybe_start=AsyncMock(), stop=AsyncMock())
        await service.join_services([s1, s2])
        s1.maybe_start.assert_awaited_once_with()
        s2.maybe_start.assert_awaited_once_with()
        s1.stop.assert_awaited_once_with()
        s2.stop.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_join_services_raises(self, *, service):
        s1 = Mock(maybe_start=AsyncMock(), stop=AsyncMock())
        s1.maybe_start.side_effect = KeyError()
        service.crash = AsyncMock()
        s2 = Mock(maybe_start=AsyncMock(), stop=AsyncMock())
        await service.join_services([s1, s2])
        s1.maybe_start.assert_awaited_once_with()
        s2.maybe_start.assert_awaited_once_with()
        s1.stop.assert_awaited_once_with()
        s2.stop.assert_awaited_once_with()

    def test_init_subclass_logger(self, *, service):
        class X(Service):
            logger = None

        class Y(X):
            logger = get_logger(__name__)
            logger.__modex__ = False

        assert Y.logger
        Y._init_subclass_logger()
        type(service).logger = None
        service._init_subclass_logger()
        assert type(service).logger.__modex__
        service._init_subclass_logger()

    def test_get_set_loop(self, *, service):
        m = Mock()
        service.loop = m
        assert service.loop is m

    def test__get_tasks__no_tasks(self, *, service):
        class X(type(service)):
            ...

        X._tasks = []
        assert list(X()._get_tasks()) == []

    @pytest.mark.asyncio
    async def test__execute_task__loop_is_closed(self, *, service):
        async def raises():
            raise RuntimeError("Event loop is closed because blah")

        await service._execute_task(raises())

    @pytest.mark.asyncio
    async def test__execute_task__exception(self, *, service):
        service.crash = AsyncMock()
        exc = KeyError("foo bah bar")

        async def raises():
            raise exc

        await service._execute_task(raises())
        service.crash.assert_awaited_once_with(exc)

    @pytest.mark.asyncio
    async def test__execute_task__RuntimeError(self, *, service):
        service.crash = AsyncMock()
        exc = RuntimeError("foo bah bar")

        async def raises():
            raise exc

        await service._execute_task(raises())
        service.crash.assert_awaited_once_with(exc)

    @pytest.mark.asyncio
    async def test__execute_task__CancelledError(self, *, service):
        async def raises():
            raise asyncio.CancelledError()

        await service._execute_task(raises())

    @pytest.mark.asyncio
    async def test__execute_task__CancelledError_stopped(self, *, service):
        service._stopped.set()

        async def raises():
            raise asyncio.CancelledError()

        await service._execute_task(raises())

    @pytest.mark.asyncio
    async def test_wait__no_args(self, *, service):
        service._wait_stopped = AsyncMock()
        assert await service.wait(timeout=1.13) == WaitResult(None, True)
        service._wait_stopped.assert_awaited_once_with(timeout=1.13)

    @pytest.mark.asyncio
    async def test_wait__one(self, *, service):
        service._wait_one = AsyncMock()
        coro = Mock()
        ret = await service.wait(coro, timeout=1.14)
        assert ret is service._wait_one.return_value
        service._wait_one.assert_awaited_once_with(coro, timeout=1.14)

    @pytest.mark.asyncio
    async def test_wait_many(self, *, service):
        with (
            patch("asyncio.wait", AsyncMock()) as wait,
            patch("asyncio.ensure_future", Mock()) as ensure_future,
        ):
            service._wait_one = AsyncMock()
            m1 = AsyncMock()
            m2 = AsyncMock()
            res = await service.wait_many([m1, m2], timeout=3.34)
            assert res is service._wait_one.return_value

            ensure_future.assert_has_calls(
                [call(m1, loop=service.loop), call(m2, loop=service.loop)]
            )
            wait.assert_called_once_with(
                [ensure_future.return_value, ensure_future.return_value],
                return_when=asyncio.ALL_COMPLETED,
                timeout=3.34,
            )

            service._wait_one.assert_awaited_once_with(ANY, timeout=3.34)

    @pytest.mark.asyncio
    async def test_wait_first__propagates_exceptions(self, *, service):
        exc = KeyError("foo")
        m1 = Mock()
        m1.done.return_value = True
        m1.exception.return_value = exc
        m1.result.side_effect = exc
        with patch("asyncio.wait", AsyncMock()) as wait:
            wait.return_value = ((m1,), ())
            with pytest.raises(KeyError):
                await service.wait_first(asyncio.sleep(5), timeout=5)

    @pytest.mark.asyncio
    async def test_wait_first__propagates_CancelledError(self, *, service):
        m1 = Mock()
        m1.done.return_value = False
        m1.cancelled.return_value = False
        sleep = asyncio.sleep(5)
        to_cancel = [sleep]
        try:
            with patch("asyncio.ensure_future") as ensure_future:
                fut = ensure_future.return_value
                fut.done.return_value = False
                fut.cancelled.return_value = True
                with patch("asyncio.wait", AsyncMock()) as wait:
                    wait.return_value = ((m1,), ())
                    with pytest.raises(asyncio.CancelledError):
                        await service.wait_first(sleep, timeout=5)

                for call_ in ensure_future.call_args_list:
                    # Service.wait()
                    # calls ensure_future(self._stopped.wait())
                    # and since we have monkeypatched ensure_future
                    # to return a mock, it will not properly clean up.
                    # we need to clean up the coroutine we created
                    awaitable_arg = call_[0][0]
                    to_cancel.append(awaitable_arg)
        finally:
            # due to the way we mock this up, we need to
            # ``await`` any dangling ``async def`` calls
            # to avoid the "coroutine x was never awaited" warnings.
            for fut in to_cancel:
                handle = asyncio.ensure_future(fut)
                handle.cancel()
                try:
                    await handle
                except (asyncio.CancelledError, RuntimeError):
                    pass

    @pytest.mark.asyncio
    async def test_wait_until_stopped(self, *, service):
        service.wait = AsyncMock()
        await service.wait_until_stopped()
        service.wait.assert_called_once_with()

    def test_service_reset__None_in_children(self, *, service):
        service._children = [None]
        service.service_reset()

    @pytest.mark.asyncio
    async def test__default_stop_children__None(self, *, service):
        service._children = [None]
        await service._default_stop_children()

    @pytest.mark.asyncio
    async def test__default_stop__raises(self, *, service):
        service.log.exception = Mock()
        service._children = [Mock(stop=AsyncMock(side_effect=KeyError()))]
        await service._default_stop_children()
        service.log.exception.assert_called_once()

    @pytest.mark.asyncio
    async def test_crash__already_crashed(self, *, service):
        service._crashed.set()
        await service.crash(KeyError())

    @pytest.mark.asyncio
    async def test_crash__recursive_loop(self, *, service):
        service._crashed.clear()
        service.log.warning = Mock()
        service.supervisor = None
        obj = Mock(data="foo", children=[Mock(), Mock(), Mock()])
        service.beacon = Mock(
            depth=3,
            walk=Mock(return_value=[obj, obj]),
        )
        await service.crash(KeyError())
        service.log.warning.assert_called_once()

    @pytest.mark.asyncio
    async def test_crash__recursive_loop__no_root(self, *, service):
        service._crashed.clear()
        service.log.warning = Mock()
        service.supervisor = None
        obj = Mock(data="foo", children=[Mock(), Mock(), Mock()])
        service.beacon = Mock(
            depth=3,
            walk=Mock(return_value=[obj, obj]),
        )
        service.beacon.root = None
        await service.crash(KeyError())
        service.log.warning.assert_called_once()

    @pytest.mark.asyncio
    async def test__gather_futures__raises_cancel(self, *, service):
        service._futures = [Mock()]
        service._maybe_wait_for_futures = AsyncMock()

        def on_wait_for_futures(**kwargs):
            if service._maybe_wait_for_futures.call_count >= 3:
                service._futures.clear()
            raise asyncio.CancelledError()

        service._maybe_wait_for_futures.side_effect = on_wait_for_futures

        await service._gather_futures()

        assert service._maybe_wait_for_futures.await_count == 3

    @pytest.mark.asyncio
    async def test__maybe_wait_for_futures__ValueError_left(self, *, service):
        service._futures = [Mock()]
        with patch("asyncio.shield", AsyncMock()) as shield:
            with patch("asyncio.wait", AsyncMock()):

                async def on_shield(fut, *args, **kwargs):
                    # if we don't wait for coroutine passed to shield
                    # we get 'was never awaited' warning.
                    await fut
                    raise ValueError()

                shield.side_effect = on_shield
                with pytest.raises(ValueError):
                    await service._maybe_wait_for_futures()

    @pytest.mark.asyncio
    async def test__maybe_wait_for_futures__ValueError_empty(self, *, service):
        service._futures = [Mock()]
        with patch("asyncio.shield", AsyncMock()) as shield:
            with patch("asyncio.wait", AsyncMock()):

                async def on_shield(fut, *args, **kwargs):
                    # if we don't wait for coroutine passed to shield
                    # we get 'was never awaited' warning.
                    await fut
                    service._futures.clear()
                    raise ValueError()

                shield.side_effect = on_shield
                await service._maybe_wait_for_futures()

    @pytest.mark.asyncio
    async def test__maybe_wait_for_futures__CancelledError(self, *, service):
        service._futures = [Mock()]
        with patch("asyncio.shield", AsyncMock()) as shield:
            with patch("asyncio.wait", AsyncMock()):

                async def on_shield(fut, *args, **kwargs):
                    # if we don't wait for coroutine passed to shield
                    # we get 'was never awaited' warning.
                    await fut
                    raise asyncio.CancelledError()

                shield.side_effect = on_shield
                await service._maybe_wait_for_futures()

    @pytest.mark.asyncio
    async def test_itertimer(self, *, service):
        with patch("mode.services.Timer") as itertimer:

            async def on_itertimer(*args, **kwargs):
                yield 1.0
                yield 1.003
                yield 0.995

            itertimer.side_effect = on_itertimer

            service.sleep = AsyncMock()
            values = [value async for value in service.itertimer(1.0)]
            assert values == [1.0, 1.003, 0.995]

    @pytest.mark.asyncio
    async def test_itertimer__first_stop(self, *, service):
        with patch("mode.services.Timer") as itertimer:

            async def on_itertimer(*args, **kwargs):
                service._stopped.set()
                yield 1.0

            itertimer.side_effect = on_itertimer
            values = [value async for value in service.itertimer(1.0)]
            assert values == []

    @pytest.mark.asyncio
    async def test_itertimer__second_stop(self, *, service):
        with patch("mode.services.Timer") as itertimer:

            async def on_itertimer(*args, **kwargs):
                for val in [0.784512, 0.2, 0.3]:
                    await service.sleep(val)
                    yield val

            itertimer.side_effect = on_itertimer

            service.sleep = AsyncMock(name="sleep")

            async def on_sleep(*args, **kwargs):
                service._stopped.set()
                return None

            service.sleep.side_effect = on_sleep
            values = [value async for value in service.itertimer(1.0)]

            assert values == []
            service.sleep.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_itertimer__third_stop(self, *, service):
        with patch("mode.services.Timer") as itertimer:

            async def on_itertimer(*args, **kwargs):
                yield 0.1341
                yield 0.2
                yield 0.3
                yield 0.4

            itertimer.side_effect = on_itertimer

            sleep = AsyncMock(name="sleep")

            values = []
            async for value in service.itertimer(1.0, sleep=sleep):
                values.append(value)
                service._stopped.set()
            assert values == [0.1341]



import typing
from typing import (
    AbstractSet,
    ClassVar,
    Dict,
    FrozenSet,
    Generic,
    List,
    Mapping,
    MutableMapping,
    MutableSet,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)
from unittest.mock import Mock, ANY

import abc
import collections.abc
import pickle
import sys

import pytest

from mode import Service, ServiceT
from mode.services import ServiceBase, ServiceCallbacks
from mode.utils.mocks import IN
from mode.utils.objects import (
    ForwardRef,
    InvalidAnnotation,
    KeywordReduce,
    Unordered,
    _ForwardRef_safe_eval,
    _remove_optional,
    _restore_from_keywords,
    canoname,
    canonshortname,
    guess_polymorphic_type,
    is_optional,
    is_union,
    iter_mro_reversed,
    label,
    qualname,
    remove_optional,
    reveal_annotations,
    shortname,
)

PY37 = sys.version_info >= (3, 7)

EXTRA_GENERIC_INHERITS_FROM = []
if PY37:
    # typing.Generic started inheriting from abc.ABC in Python 3.7.0,
    # so this is needed for iter_mro_reversed test below.
    EXTRA_GENERIC_INHERITS_FROM = [abc.ABC]


class D(Service):
    ...


class C(D):
    ...


class B(C):
    ...


class A(B):
    ...


@pytest.mark.parametrize(
    "cls,stop,expected_mro",
    [
        (A, Service, [D, C, B, A]),
        (B, Service, [D, C, B]),
        (C, Service, [D, C]),
        (D, Service, [D]),
        (
            A,
            object,
            (
                [ServiceCallbacks, Generic]
                + EXTRA_GENERIC_INHERITS_FROM
                + [ANY, ServiceT, ServiceBase, Service, D, C, B, A]
            ),
        ),
        (A, B, [A]),
        (A, C, [B, A]),
        (A, D, [C, B, A]),
    ],
)
def test_iter_mro_reversed(cls, stop, expected_mro):
    assert list(iter_mro_reversed(cls, stop=stop)) == expected_mro


def test_Unordered():

    assert Unordered(1) < Unordered(10)
    x = set()
    x.add(Unordered({"foo": "bar"}))
    x.add(Unordered({"foo": "bar"}))
    assert len(x) == 2
    assert repr(x)


def test__restore_from_keywords():
    m = Mock()
    _restore_from_keywords(m, {"foo": 1, "bar": 20})
    m.assert_called_once_with(foo=1, bar=20)


class X(KeywordReduce):
    def __init__(self, name, age):
        self.name = name
        self.age = age

    def __reduce_keywords__(self):
        return {"name": self.name, "age": self.age}


def test_KeywordReduce():
    with pytest.raises(NotImplementedError):
        KeywordReduce().__reduce_keywords__()

    x = X("foo", 10)
    y = pickle.loads(pickle.dumps(x))
    assert y.name == x.name
    assert y.age == x.age


def test_qualname_object():
    class X:
        ...

    assert qualname("foo") == "builtins.str"
    assert qualname(str) == "builtins.str"

    assert qualname(X).endswith("test_qualname_object.<locals>.X")
    assert qualname(X()).endswith("test_qualname_object.<locals>.X")


def test_shortname_object():
    class X:
        ...

    assert shortname("foo") == "builtins.str"
    assert shortname(str) == "builtins.str"

    assert shortname(X) == __name__ + ".X"
    assert shortname(X()) == __name__ + ".X"


def test_canoname():
    class X:
        ...

    X.__module__ = "__main__"
    x = X()

    class Y:
        ...

    y = Y()

    assert canoname(X, main_name="faust") == "faust.test_canoname.<locals>.X"
    assert canoname(x, main_name="faust") == "faust.test_canoname.<locals>.X"
    assert canoname(Y, main_name="faust") == ".".join(
        [
            __name__,
            "test_canoname.<locals>.Y",
        ]
    )
    assert canoname(y, main_name="faust") == ".".join(
        [
            __name__,
            "test_canoname.<locals>.Y",
        ]
    )


def test_canonshortname():
    class X:
        ...

    X.__module__ = "__main__"
    x = X()

    class Y:
        ...

    y = Y()

    assert canonshortname(X, main_name="faust") == "faust.X"
    assert canonshortname(x, main_name="faust") == "faust.X"
    assert canonshortname(Y, main_name="faust") == ".".join(
        [
            __name__,
            "Y",
        ]
    )
    assert canonshortname(y, main_name="faust") == ".".join(
        [
            __name__,
            "Y",
        ]
    )


def test_annotations():
    class X:
        Foo: ClassVar[int] = 3
        foo: "int"
        bar: List["X"]
        baz: Union[List["X"], str]
        mas: int = 3

    fields, defaults = reveal_annotations(
        X,
        globalns=globals(),
        localns=locals(),
    )

    assert fields == {
        "Foo": ClassVar[int],
        "foo": int,
        "bar": List[X],
        "baz": Union[List[X], str],
        "mas": int,
    }
    assert defaults["mas"] == 3


def test_annotations__skip_classvar():
    class X:
        Foo: ClassVar[int] = 3
        foo: "int"
        bar: List["X"]
        baz: Union[List["X"], str]
        mas: int = 3

    fields, defaults = reveal_annotations(
        X,
        globalns=globals(),
        localns=locals(),
        skip_classvar=True,
    )

    assert fields == {
        "foo": int,
        "bar": List[X],
        "baz": Union[List[X], str],
        "mas": int,
    }
    assert defaults["mas"] == 3


def test_annotations__invalid_type():
    class X:
        foo: List

    with pytest.raises(InvalidAnnotation):
        reveal_annotations(
            X,
            globalns=globals(),
            localns=locals(),
            invalid_types={List},
            skip_classvar=True,
        )


def test_annotations__no_local_ns_raises():
    class Bar:
        ...

    class X:
        bar: "Bar"

    with pytest.raises(NameError):
        reveal_annotations(
            X,
            globalns=None,
            localns=None,
        )


def test__ForwardRef_safe_eval():
    ref1 = ForwardRef("int")
    assert _ForwardRef_safe_eval(ref1) == int
    assert _ForwardRef_safe_eval(ref1) == int
    assert ref1.__forward_evaluated__
    assert ref1.__forward_value__ == int
    assert _ForwardRef_safe_eval(ForwardRef("foo"), localns={"foo": str}) == str
    assert _ForwardRef_safe_eval(
        ForwardRef("ClassVar[int]"), globalns=globals(), localns=locals()
    )


# Union[type(None)] actually returns None
# so we have to construct this object to test condition in code.
WeirdNoneUnion = Union[str, int]
WeirdNoneUnion.__args__ = (type(None), type(None))


@pytest.mark.parametrize(
    "input,expected",
    [
        (Optional[str], str),
        (Union[str, None], str),
        (Union[str, type(None)], str),
        (Union[str, None], str),
        (Optional[List[str]], List[str]),
        (Optional[Mapping[int, str]], Mapping[int, str]),
        (Optional[AbstractSet[int]], AbstractSet[int]),
        (Optional[Set[int]], Set[int]),
        (Optional[Tuple[int, ...]], Tuple[int, ...]),
        (Optional[Dict[int, str]], Dict[int, str]),
        (Optional[List[int]], List[int]),
        (str, str),
        (List[str], List[str]),
        (Union[str, int, float], Union[str, int, float]),
        (WeirdNoneUnion, WeirdNoneUnion),
    ],
)
def test_remove_optional(input, expected):
    assert remove_optional(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (Optional[str], ((), str)),
        (Union[str, None], ((), str)),
        (Union[str, type(None)], ((), str)),
        (Union[str, None], ((), str)),
        (Optional[List[str]], ((str,), list)),
        (
            Optional[Mapping[int, str]],
            ((int, str), IN(dict, collections.abc.Mapping, typing.Mapping)),
        ),
        (
            Optional[AbstractSet[int]],
            ((int,), IN(set, collections.abc.Set, typing.AbstractSet)),
        ),
        (
            Optional[Set[int]],
            ((int,), IN(set, collections.abc.Set, typing.AbstractSet)),
        ),
        (Optional[Tuple[int, ...]], ((int, ...), IN(tuple, typing.Tuple))),
        (Optional[Dict[int, str]], ((int, str), dict)),
        (Optional[List[int]], ((int,), list)),
        (str, ((), str)),
        (List[str], ((str,), list)),
        (WeirdNoneUnion, ((type(None), type(None)), Union)),
    ],
)
def test__remove_optional__find_origin(input, expected):
    assert _remove_optional(input, find_origin=True) == expected


def test__remove_optional_edgecase():
    input = Union[str, int, float]
    expected = (str, int, float)
    res = _remove_optional(input, find_origin=True)
    assert res[0] == expected
    # must use `is` here on Python 3.6
    assert res[1] is typing.Union


@pytest.mark.parametrize(
    "input,expected",
    [
        (Optional[str], True),
        (Union[str, None], True),
        (Union[str, type(None)], True),
        (Union[str, None], True),
        (str, False),
        (List[str], False),
        (Union[str, int, float], False),
    ],
)
def test_is_optional(input, expected):
    assert is_optional(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (Tuple[int, ...], (tuple, int)),
        (List[int], (list, int)),
        (Mapping[str, int], (dict, int)),
        (Dict[str, int], (dict, int)),
        (MutableMapping[str, int], (dict, int)),
        (Set[str], (set, str)),
        (FrozenSet[str], (set, str)),
        (MutableSet[str], (set, str)),
        (AbstractSet[str], (set, str)),
        (Sequence[str], (list, str)),
    ],
)
def test_guess_polymorphic_type(input, expected):
    assert guess_polymorphic_type(input) == expected
    assert guess_polymorphic_type(Optional[input]) == expected
    assert guess_polymorphic_type(Union[input, None]) == expected


def test_guess_polymorphic_type__not_generic():
    class X:
        ...

    with pytest.raises(TypeError):
        guess_polymorphic_type(str)
    with pytest.raises(TypeError):
        guess_polymorphic_type(bytes)
    with pytest.raises(TypeError):
        guess_polymorphic_type(X)


def test_label_pass():
    s = "foo"
    assert label(s) is s


@pytest.mark.parametrize(
    "input,expected",
    [
        (str, False),
        (int, False),
        (Union[int, bytes], True),
        (Optional[str], True),
    ],
)
def test_is_union(input, expected):
    assert is_union(input) == expected



from typing import IO, Type

import asyncio
import io
import logging
import sys
import time
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from unittest.mock import ANY, AsyncMock, Mock, call, patch

import pytest

from mode.utils.logging import (
    LOG_RECORD_BUILTINS,
    CompositeLogger,
    DefaultFormatter,
    ExtensionFormatter,
    FileLogProxy,
    LogMessage,
    Logwrapped,
    _FlightRecorderProxy,
    _formatter_registry,
    _setup_logging,
    current_flight_recorder,
    flight_recorder,
    formatter,
    get_logger,
    level_name,
    level_number,
    on_timeout,
    print_task_name,
    redirect_stdouts,
    setup_logging,
)


def log_called_with(logger, *args, stacklevel, **kwargs):
    logger.log.assert_called_once_with(*args, stacklevel=stacklevel, **kwargs)


def formatter_called_with(formatter, *args, stacklevel, **kwargs):
    formatter.assert_called_once_with(*args, stacklevel=stacklevel, **kwargs)


def test_log_record_builtins():
    rv = logging.makeLogRecord({})
    assert LOG_RECORD_BUILTINS - {"message", "asctime", "extra"} == set(
        rv.__dict__.keys()
    )


class test_CompositeLogger:
    @pytest.fixture()
    def logger(self):
        return Mock(name="logger")

    @pytest.fixture()
    def formatter(self):
        return Mock(name="formatter")

    @pytest.fixture()
    def log(self, *, logger, formatter):
        return CompositeLogger(logger=logger, formatter=formatter)

    def test_log(self, *, log, logger, formatter):
        log.log(logging.INFO, "msg", 1, kw=2)
        log_called_with(
            logger,
            logging.INFO,
            formatter.return_value,
            1,
            stacklevel=2,
            kw=2,
        )
        formatter_called_with(formatter, logging.INFO, "msg", 1, kw=2, stacklevel=2)

    def test_log__no_formatter(self, *, log, logger):
        log.formatter = None
        log.log(logging.INFO, "msg", 1, kw=2)
        log_called_with(
            logger,
            logging.INFO,
            "msg",
            1,
            kw=2,
            stacklevel=2,
        )

    @pytest.mark.parametrize(
        "method,severity,extra",
        [
            ("debug", logging.DEBUG, {}),
            ("info", logging.INFO, {}),
            ("warn", logging.WARN, {}),
            ("warning", logging.WARN, {}),
            ("error", logging.ERROR, {}),
            ("crit", logging.CRITICAL, {}),
            ("critical", logging.CRITICAL, {}),
            ("exception", logging.ERROR, {"exc_info": 1}),
        ],
    )
    def test_severity_mixin(self, method, severity, extra, *, log, logger):
        log.formatter = None
        getattr(log, method)("msg", "arg1", kw1=3, kw2=5)
        log_called_with(
            logger,
            severity,
            "msg",
            "arg1",
            kw1=3,
            kw2=5,
            stacklevel=3,
            **extra,
        )

    def test_dev__enabled(self, log):
        log.log = Mock()
        with patch("mode.utils.logging.DEVLOG", True):
            log.dev("msg", 1, k=2)
            log_called_with(log, logging.INFO, "msg", 1, k=2, stacklevel=3)

    def test_dev__disabled(self, log):
        log.info = Mock()
        with patch("mode.utils.logging.DEVLOG", False):
            log.dev("msg", 1, k=2)
            log.info.assert_not_called()


def test_formatter():
    f = Mock()
    formatter(f)
    try:
        assert f in _formatter_registry
    finally:
        _formatter_registry.remove(f)


@pytest.mark.parametrize("MyFormatter", (DefaultFormatter, ExtensionFormatter))
class test_Formatter:
    def test_format(self, MyFormatter: Type[logging.Formatter]):
        extra = {"moo": 30, "baz": 1111, "foo": "hello"}
        record = logging.getLogger("__test_Formatter__").makeRecord(
            "name",
            logging.INFO,
            "path",
            303,
            r"msg %s",
            (1,),
            None,
            extra=extra,
        )
        _ = MyFormatter().format(record)
        keys = set(s.split("=")[0] for s in record.extra.split(", "))
        assert keys == set(extra.keys())

    def test_formatTime(self, MyFormatter):
        record = logging.LogRecord(
            "name",
            logging.INFO,
            "path",
            303,
            "msg",
            {"foo": 1, "extra": {"data": {"moo": 30, "baz": [1, 2]}}},
            exc_info=None,
        )
        s = MyFormatter().formatTime(record)
        iso = datetime.fromtimestamp(
            record.created, tz=timezone(timedelta(seconds=-time.timezone))
        ).isoformat(timespec="milliseconds")
        assert s == iso


@pytest.mark.parametrize(
    "input,expected",
    [
        ("DEBUG", logging.DEBUG),
        ("INFO", logging.INFO),
        ("WARNING", logging.WARNING),
        ("WARNING", logging.WARNING),
        ("ERROR", logging.ERROR),
        ("CRITICAL", logging.CRITICAL),
        (logging.ERROR, logging.ERROR),
    ],
)
def test_level_number(input, expected):
    assert level_number(input) == expected


@pytest.mark.parametrize(
    "input",
    (None, 10.0, 0.0),
)
def test_level_number__type(input):
    with pytest.raises(TypeError):
        level_number(input)


@pytest.mark.parametrize(
    "input,expected",
    [
        (logging.DEBUG, "DEBUG"),
        (logging.INFO, "INFO"),
        (logging.WARN, "WARNING"),
        (logging.WARNING, "WARNING"),
        (logging.ERROR, "ERROR"),
        (logging.CRITICAL, "CRITICAL"),
        ("INFO", "INFO"),
    ],
)
def test_level_name(input, expected):
    assert level_name(input) == expected


@pytest.mark.parametrize(
    "input",
    (None, 10.0, 0.0),
)
def test_level_name__type(input):
    with pytest.raises(TypeError):
        level_name(input)


class test_setup_logging:
    def test_default(self):
        with patch("mode.utils.logging._setup_logging") as _sl:
            setup_logging(log_level="INFO", log_file=None)

            _sl.assert_called_once_with(
                level=logging.INFO,
                filename=None,
                stream=sys.stdout,
                log_handlers=None,
                logging_config=None,
            )

    def test_logfile(self):
        with patch("mode.utils.logging._setup_logging") as _sl:
            setup_logging(log_level="INFO", log_file="foo.txt")

            _sl.assert_called_once_with(
                level=logging.INFO,
                filename="foo.txt",
                stream=None,
                log_handlers=None,
                logging_config=None,
            )

    def test_io(self):
        logfile = Mock(spec=IO)
        with patch("mode.utils.logging._setup_logging") as _sl:
            setup_logging(log_level="INFO", log_file=logfile)

            _sl.assert_called_once_with(
                level=logging.INFO,
                filename=None,
                stream=logfile,
                log_handlers=None,
                logging_config=None,
            )

    def test_io_no_tty(self):
        logfile = Mock(spec=IO)
        logfile.isatty.side_effect = AttributeError()
        with patch("mode.utils.logging._setup_logging") as _sl:
            setup_logging(log_level="INFO", log_file=logfile)

            _sl.assert_called_once_with(
                level=logging.INFO,
                filename=None,
                stream=logfile,
                log_handlers=None,
                logging_config=None,
            )


class test__setup_logging:
    def setup_method(self, method):
        self.extension_formatter_patch = patch("mode.utils.logging.ExtensionFormatter")
        self.extension_formatter = self.extension_formatter_patch.start()
        self.colorlog_patch = patch("mode.utils.logging.colorlog")
        self.colorlog = self.colorlog_patch.start()
        self.logging_patch = patch("mode.utils.logging.logging")
        self.logging = self.logging_patch.start()

    def teardown_method(self):
        self.extension_formatter_patch.stop()
        self.colorlog_patch.stop()
        self.logging_patch.stop()

    def test_get_logger(self):
        assert get_logger(__name__)
        assert get_logger(__name__).handlers == get_logger(__name__).handlers

    def test_setup_logging_helper_both_filename_and_stream(self):
        with pytest.raises(AssertionError):
            _setup_logging(filename="TEMP", stream=Mock())

    def test_setup_logging_helper_with_filename(self):
        _setup_logging(filename="TEMP")
        self.logging.config.dictConfig.assert_called_once_with(ANY)

    def test_setup_logging_helper_with_stream_no_handlers(self):
        _setup_logging(stream=Mock())
        self.logging.config.dictConfig.assert_called_once_with(ANY)

    def test_setup_logging_helper_with_stream(self):
        mock_handler = Mock()
        _setup_logging(
            filename=None,
            stream=Mock(),
            log_handlers=[mock_handler],
        )
        self.logging.config.dictConfig.assert_called_once_with(ANY)
        self.logging.root.addHandler.assert_called_once_with(mock_handler)

    def test_setup_logging_helper_with_merge_config(self):
        _setup_logging(
            filename=None,
            stream=Mock(),
            logging_config={"merge": True, "foo": 1},
        )
        self.logging.config.dictConfig.assert_called_once_with(ANY)

    def test_setup_logging_helper_no_merge_config(self):
        _setup_logging(
            logging_config={"merge": False, "foo": 1},
        )
        self.logging.config.dictConfig.assert_called_once_with(ANY)


class test_Logwrapped:
    @pytest.fixture()
    def obj(self):
        return Mock(name="obj")

    @pytest.fixture()
    def logger(self):
        return Mock(name="logger")

    @pytest.fixture()
    def wrapped(self, *, obj, logger):
        return Logwrapped(obj, logger, severity="INFO", ident="ident")

    def test_constructor(self, wrapped):
        assert wrapped.severity == logging.INFO

    def test_wrapper(self, wrapped, obj):
        obj.calculate.__name__ = "calculate"
        wrapped.calculate(1, 2, kw=1)
        obj.calculate.assert_called_once_with(1, 2, kw=1)

    def test_wrapper__no_ident(self, wrapped, obj):
        wrapped.ident = None
        obj.calculate.__name__ = "calculate"
        wrapped.calculate(1, 2, kw=1)
        obj.calculate.assert_called_once_with(1, 2, kw=1)

    def test_wrapper__no_args(self, wrapped, obj):
        obj.calculate.__name__ = "calculate"
        wrapped.calculate()
        obj.calculate.assert_called_once_with()

    def test_wrapper__only_kwargs(self, wrapped, obj):
        obj.calculate.__name__ = "calculate"
        wrapped.calculate(kw=3)
        obj.calculate.assert_called_once_with(kw=3)

    def test_wrapper__ignored(self, wrapped, logger, obj):
        logger.log.assert_not_called()
        obj.__enter__ = Mock()
        obj.__exit__ = Mock()
        assert wrapped.__enter__ is obj.__enter__
        assert wrapped.__exit__ is obj.__exit__

    def test_repr(self, wrapped, obj):
        assert repr(wrapped) == repr(obj)

    def test_dir(self, wrapped, obj):
        assert dir(wrapped) == dir(obj)


def test_print_task_name():
    out = io.StringIO()
    task = Mock()
    task.__wrapped__ = Mock()
    print_task_name(task, file=out)
    assert out.getvalue()

    task.__wrapped__ = None
    print_task_name(task, file=out)
    assert out.getvalue()


class test_flight_recorder:
    @pytest.fixture()
    def logger(self):
        return Mock(name="logger")

    @pytest.fixture()
    def bb(self, *, logger):
        return flight_recorder(logger, timeout=30.0)

    def test_wrap_debug(self, bb):
        obj = Mock()
        bb.wrap = Mock()
        bb.wrap_debug(obj)
        bb.wrap.assert_called_once_with(logging.DEBUG, obj)

    def test_wrap_info(self, bb):
        obj = Mock()
        bb.wrap = Mock()
        bb.wrap_info(obj)
        bb.wrap.assert_called_once_with(logging.INFO, obj)

    def test_wrap_warn(self, bb):
        obj = Mock()
        bb.wrap = Mock()
        bb.wrap_warn(obj)
        bb.wrap.assert_called_once_with(logging.WARN, obj)

    def test_wrap_error(self, bb):
        obj = Mock()
        bb.wrap = Mock()
        bb.wrap_error(obj)
        bb.wrap.assert_called_once_with(logging.ERROR, obj)

    def test_wrap(self, bb):
        obj = Mock()
        with patch("mode.utils.logging.Logwrapped") as Logwrapped:
            ret = bb.wrap(logging.ERROR, obj)
            assert ret is Logwrapped.return_value
            Logwrapped.assert_called_once_with(
                logger=bb,
                severity=logging.ERROR,
                obj=obj,
            )

    def test_activate(self, bb):
        bb._fut = None
        bb._waiting = Mock()
        with patch("mode.utils.logging.current_task") as current_task:
            with patch("asyncio.ensure_future") as ensure_future:
                bb.activate()
                assert bb.started_at_date
                assert bb.enabled_by is current_task.return_value
                ensure_future.assert_called_once_with(
                    bb._waiting.return_value,
                    loop=bb.loop,
                )
                assert bb._fut is ensure_future.return_value

    def test_activate__already_activated(self, bb):
        bb._fut = Mock()
        with pytest.raises(RuntimeError):
            bb.activate()

    def test_cancel(self, bb):
        bb._fut = None
        bb._logs = [1, 2, 3]
        bb.cancel()
        assert bb._logs == []
        assert bb._fut is None

        fut = bb._fut = Mock()
        bb.cancel()
        assert bb._fut is None
        fut.cancel.assert_called_once_with()  # type: ignore

    def test_log__active(self, bb, logger):
        bb._fut = Mock()
        bb._buffer_log = Mock()
        bb.log(logging.DEBUG, "msg %r %(foo)s", 1, foo="bar")
        bb._buffer_log.assert_called_once_with(
            logging.DEBUG,
            "msg %r %(foo)s",
            (1,),
            {"foo": "bar"},
        )

    def test_log__inactive(self, bb, logger):
        bb._fut = None
        bb._buffer_log = Mock()
        bb.log(logging.DEBUG, "msg %r %(foo)s", 1, foo="bar")
        log_called_with(
            logger,
            logging.DEBUG,
            "msg %r %(foo)s",
            1,
            foo="bar",
            stacklevel=2,
        )

    def test__buffer_log(self, bb):
        with patch("time.asctime") as asctime:
            bb._buffer_log(logging.ERROR, "msg %r %(foo)s", (1,), {"foo": "bar"})
            assert bb._logs[-1] == LogMessage(
                logging.ERROR,
                "msg %r %(foo)s",
                asctime(),
                (1,),
                {"foo": "bar"},
            )

    @pytest.mark.asyncio
    async def test__waiting__cancelled(self, bb):
        assert not bb._logs
        bb._buffer_log(logging.ERROR, "msg %r %(foo)s", (1,), {"foo": "bar"})
        with patch("asyncio.sleep", AsyncMock()) as sleep:
            sleep.side_effect = asyncio.CancelledError()
            await bb._waiting()
            sleep.assert_called_once_with(bb.timeout)
            assert bb._logs

    @pytest.mark.asyncio
    async def test__waiting__has_logs(self, bb):
        assert not bb._logs
        bb._buffer_log(logging.ERROR, "msg %r %(foo)s", (1,), {"foo": "bar"})
        assert bb._logs
        with patch("asyncio.sleep", AsyncMock()):
            await bb._waiting()

    @pytest.mark.asyncio
    async def test__waiting__no_logs(self, bb):
        assert not bb._logs
        with patch("asyncio.sleep", AsyncMock()):
            await bb._waiting()

    @pytest.mark.asyncio
    async def test__waiting__enabled_by(self, bb):
        assert not bb._logs
        bb.enabled_by = Mock()
        with patch("asyncio.sleep", AsyncMock()):
            with patch("mode.utils.logging.format_task_stack") as fts:
                await bb._waiting()
            fts.assert_called_once_with(bb.enabled_by)

    @pytest.mark.asyncio
    async def test__waiting__raises(self, bb):
        assert not bb._logs
        bb.logger.warning = Mock(side_effect=KeyError())
        with patch("asyncio.sleep", AsyncMock()):
            with pytest.raises(KeyError):
                await bb._waiting()

    def test_repr(self, bb):
        assert repr(bb)

    def test_context(self, bb):
        bb.activate = Mock()
        bb.cancel = Mock()
        with bb:
            bb.activate.assert_called_once_with()
        bb.cancel.assert_called_once_with()


class test_FileLogProxy:
    def test_constructor__defaults(self):
        logger = get_logger("foo")
        logger.level = None
        assert logger.level is None
        f = FileLogProxy(logger)
        assert f.severity == logging.WARN

    def test_constructor__severity_from_logger(self):
        logger = get_logger("foo")
        logger.level = logging.DEBUG
        f = FileLogProxy(logger)
        assert f.severity == logging.DEBUG

    def test_constructor__explicit_severity(self):
        logger = get_logger("foo")
        logger.level = logging.DEBUG
        f = FileLogProxy(logger, severity=logging.ERROR)
        assert f.severity == logging.ERROR

    def test__safewrap_handler(self):
        f = FileLogProxy(get_logger("foo"))
        handler = Mock()
        f._safewrap_handler(handler)

        with patch("traceback.print_exc") as print_exc:
            record = Mock()
            handler.handleError(record)
            print_exc.assert_called_once_with(None, sys.__stderr__)
            print_exc.side_effect = IOError()
            handler.handleError(record)

    def test_write(self):
        logger = Mock(handlers=[])
        f = FileLogProxy(logger)
        f._threadlocal.recurse_protection = True
        f.write("foo")
        logger.log.assert_not_called()

        f._threadlocal.recurse_protection = False
        f.write("")
        f.write("               ")
        f.close()
        f.write("msg")
        logger.log.assert_not_called()

        f._closed = False
        f.write(" msg ")
        logger.log.assert_called_once_with(f.severity, "msg")

        f.writelines(["foo", "bar"])
        logger.log.assert_has_calls(
            [
                call(f.severity, "msg"),
                call(f.severity, "foo"),
                call(f.severity, "bar"),
            ]
        )

    def test_flush(self):
        FileLogProxy(get_logger("foo")).flush()

    def test_isatty(self):
        assert not FileLogProxy(get_logger("foo")).isatty()


def test_redirect_stdouts():
    prev_stdout = sys.stdout
    prev_stderr = sys.stderr
    with redirect_stdouts(stdout=False, stderr=False):
        assert sys.stdout == prev_stdout
        assert sys.stderr == prev_stderr
    prev_stdout = sys.stdout  # pytest keep changing this
    prev_stderr = sys.stderr
    with redirect_stdouts(stdout=True, stderr=False):
        assert isinstance(sys.stdout, FileLogProxy)
        assert sys.stderr == prev_stderr
    prev_stdout = sys.stdout
    prev_stderr = sys.stderr
    with redirect_stdouts(stdout=False, stderr=True):
        assert sys.stdout == prev_stdout
        assert isinstance(sys.stderr, FileLogProxy)
    prev_stdout = sys.stdout
    prev_stderr = sys.stderr
    with redirect_stdouts(stdout=True, stderr=True):
        assert isinstance(sys.stdout, FileLogProxy)
        assert isinstance(sys.stderr, FileLogProxy)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "extra_context",
    [
        {},
        {"foo": "bar"},
    ],
)
async def test_on_timeout(extra_context):
    logger = Mock()
    assert isinstance(on_timeout, _FlightRecorderProxy)

    # Test no errors when there's no active flight recorder
    _assert_log_severities(on_timeout)

    with patch("time.asctime") as asctime:
        asctime.return_value = "TIME"
        # Test logging to active flight recorder (with nesting)
        with flight_recorder(logger, timeout=300) as fl1:
            fl1.extra_context.update(extra_context)
            assert current_flight_recorder() is fl1
            _assert_recorder_exercised(on_timeout, fl1)

            with flight_recorder(logger, timeout=30) as fl2:
                for k, v in fl1.extra_context.items():
                    assert fl2.extra_context[k] == v
                assert current_flight_recorder() is fl2
                _assert_recorder_exercised(on_timeout, fl2)
                _assert_recorder_flush_logs(logger, fl2)

            assert current_flight_recorder() is fl1
            _assert_recorder_flush_logs(logger, fl1)
            _assert_recorder_exercised(on_timeout, fl1)
            _assert_recorder_flush_logs(logger, fl1)


def _assert_log_severities(logger):
    logger.debug("DEBUG %d %(a)s", 1, a="A")
    logger.info("INFO %d %(b)s", 2, b="B")
    logger.warning("WARNING %d %(c)s", 3, c="C")
    logger.error("ERROR %d %(d)s", 4, d="D")
    logger.critical("CRITICAL %d %(e)s", 5, e="E")


def _log_kwargs(kwargs):
    kwargs.setdefault("stacklevel", 3)
    return kwargs


EXPECTED_LOG_MESSAGES = [
    LogMessage(logging.DEBUG, "DEBUG %d %(a)s", "TIME", (1,), _log_kwargs({"a": "A"})),
    LogMessage(logging.INFO, "INFO %d %(b)s", "TIME", (2,), _log_kwargs({"b": "B"})),
    LogMessage(
        logging.WARNING,
        "WARNING %d %(c)s",
        "TIME",
        (3,),
        _log_kwargs({"c": "C"}),
    ),
    LogMessage(logging.ERROR, "ERROR %d %(d)s", "TIME", (4,), _log_kwargs({"d": "D"})),
    LogMessage(
        logging.CRITICAL,
        "CRITICAL %d %(e)s",
        "TIME",
        (5,),
        _log_kwargs({"e": "E"}),
    ),
]


def _assert_recorder_exercised(logger, fl):
    _assert_log_severities(logger)
    assert fl._logs == EXPECTED_LOG_MESSAGES


def _assert_recorder_flush_logs(logger, fl):
    fl.flush_logs(ident="IDENT")

    def _get_call(sev, msg, datestr, args, kwargs):
        kw = deepcopy(kwargs)
        if fl.extra_context:
            extra = kw.setdefault("extra", {})
            data = extra.setdefault("data", {})
            data.update(fl.extra_context)
        return call(sev, f"[%s] (%s) {msg}", "IDENT", datestr, *args, **kw)

    logger.log.assert_has_calls(
        _get_call(sev, msg, datestr, args, kwargs)
        for sev, msg, datestr, args, kwargs in EXPECTED_LOG_MESSAGES
    )
